/******************************************************************************
*
* (c) 2006 by BECK IPC GmbH
*
*******************************************************************************
*
* Disclaimer: This program is an example and should be used as such.
*             If you wish to use this program or parts of it in your application,
*             you must validate the code yourself. BECK IPC GmbH can not be held
*             responsible for the correct functioning or coding of this example.
*
* Module:     fdisk.c
*
* Function:   Simple fdisk implementation for @CHIP-RTOS.
*
*******************************************************************************
*
* $Header: c:\prjPib\IPC@CHIP\Examples\extdisk\FDISK\Source\fdisk.c, 13, 05.01.2007 08:37:53, Jan Schatz$
*
******************************************************************************/



#include <stdlib.h>
#include <clib.h>
#include <mem.h>
#include "bool.h"
#include "fdisk.h"
#include "extlib.h"



/******************************************************************************
* Global variables
******************************************************************************/
static TMasterBootRecord _mbr;      // Structure to keep the MBR data
static TMediaData _mediaData;       // Structure to keep disk media data
static unsigned char _buffer[512];  // Buffer with space for one sector
static unsigned char _drive = 0;    // Drive to operate on



/******************************************************************************
* fdiskCheckInterfaceVersion()
*
* Checks if the Ext IDE drivers interface version is >= 0x01
*
* Parameters:
*  None
*
* Result:
*  TRUE  - On success.
*  FALSE - If version is < 0x01 or unknown.
******************************************************************************/
FdiskBool fdiskCheckInterfaceVersion( void )
{
  return ( extlibGetInterfaceVersion( _drive ) >= 0x01 ) ? TRUE : FALSE;
}



/******************************************************************************
* fdiskPhysicToLogic()
*
* Converts physical cylinder-head-sector address to an logical sector.
*
* Parameters:
*  cylinder -
*  head     -
*  sector   - CHS value to convert to logical sector
*
* Result:
*  Logical sector
******************************************************************************/
unsigned long fdiskPhysicToLogic( unsigned short cylinder,
                                  unsigned short head,
                                  unsigned short sector )
{
  // Calculate logical sector
  return (unsigned long)cylinder * _mediaData.nHeads * _mediaData.nSectors +
         head * _mediaData.nSectors +
         sector - 1;  // Sector counting starts at 1
}



/******************************************************************************
* fdiskLogicToPhysic()
*
* Converts logical sector to physical cylinder-head-sector address.
*
* Parameters:
*  lSector  - Logical sector to convert
*  cylinder -
*  head     -
*  sector   - Result parameters containing the resulting CHS value
*
* Result:
*  None
******************************************************************************/
void fdiskLogicToPhysic( unsigned long lSector,
                         unsigned short *cylinder,
                         unsigned short *head,
                         unsigned short *sector )
{
  // Calculate cylinder, head and sector
  *cylinder = (unsigned short)( lSector / ( (unsigned long)_mediaData.nHeads * _mediaData.nSectors ) );
  lSector %= (unsigned long)_mediaData.nHeads * _mediaData.nSectors;
  *head = (unsigned short)( lSector / _mediaData.nSectors );
  *sector = (unsigned short)( lSector % _mediaData.nSectors )
            + 1;  // Sector Counting starts at 1
}



/******************************************************************************
* fdiskReadMbr()
*
* Reads the Master Boot Record from the disk and stores it in the global
* variable _mbr.
*
* Parameters:
*  drive - Drive to operate on. This drive will also be used for all following
*          operations. Valid values are 'B' and 'D'.
*
* Result
*  TRUE  - On success.
*  FALSE - If Master Boot Record could not be read from disk.
******************************************************************************/
FdiskBool fdiskReadMbr( unsigned char drive )
{
  unsigned short i,       // Counter variable
                 offset;  // Offset of partition table entry


  // Check parameter
  if( drive != 'B' &&
      drive != 'D' )
    return FALSE;

  // Check for correct interface version
  if( !fdiskCheckInterfaceVersion() )
    return FALSE;

  // Save drive letter in global variable
  _drive = drive;

  // Get media data
  // Disk translation (CHS) is not checked here. The EXTDISK driver has to care
  // for it However, if a partition is to be created with CHS values that exceed
  // { 1023, 255, 63 }, all values will be set to their maximum and the logical
  // addresses must be used when mounting the partition
  extlibGetMediaData( _drive, &_mediaData );


  // Read first sector of disk
  if( !extlibRead( _drive, 0, 1, _buffer ) )
  {
    return FALSE;
  }

  // Fill _mbr
  memcpy( &_mbr, _buffer, 0x01BE );

  // Partition table entries
  for( i = 0; i < 4; ++i )
  {
    offset = 0x01BE + i * PARTITION_ENTRY_SIZE;

    _mbr.partition[i].bootState = _buffer[offset+0];

    _mbr.partition[i].startHead = _buffer[offset+1];
    _mbr.partition[i].startSector = _buffer[offset+2] & 0x3F;
    _mbr.partition[i].startCylinder = (unsigned short)_buffer[offset+3];
    _mbr.partition[i].startCylinder += (unsigned short)( _buffer[offset+2] & 0xC0 ) << 2;

    _mbr.partition[i].typeId = _buffer[offset+4];

    _mbr.partition[i].endHead = _buffer[offset+5];
    _mbr.partition[i].endSector = _buffer[offset+6] & 0x3F;
    _mbr.partition[i].endCylinder = (unsigned short)_buffer[offset+7];
    _mbr.partition[i].endCylinder += (unsigned short)( _buffer[offset+6] & 0xC0 ) << 2;

    _mbr.partition[i].startLSector = (unsigned long)_buffer[offset+8];
    _mbr.partition[i].startLSector += (unsigned long)_buffer[offset+9] << 8;
    _mbr.partition[i].startLSector += (unsigned long)_buffer[offset+10] << 16;
    _mbr.partition[i].startLSector += (unsigned long)_buffer[offset+11] << 24;

    _mbr.partition[i].nSectors = (unsigned long)_buffer[offset+12];
    _mbr.partition[i].nSectors += (unsigned long)_buffer[offset+13] << 8;
    _mbr.partition[i].nSectors += (unsigned long)_buffer[offset+14] << 16;
    _mbr.partition[i].nSectors += (unsigned long)_buffer[offset+15] << 24;
  }

  _mbr.mbrSignature[0] = _buffer[510];
  _mbr.mbrSignature[1] = _buffer[511];

  return TRUE;
}



/******************************************************************************
* fdiskCheckMbr()
*
* Checks the MBR structure in the global variable _mbr.
*
* Parameters:
*  None
*
* Result
*  TRUE  - On success.
*  FALSE - If Master Boot Record is invalid or unusable for RTOS.
******************************************************************************/
FdiskBool fdiskCheckMbr( void )
{
  short i, j;               // Counter variable
  FdiskBool result = TRUE;  // Return value


  // Has signature?
  if( _mbr.mbrSignature[0] != 0x55 || _mbr.mbrSignature[1] != 0xAA )
  {
    helper_strout( "\r\nWarning: MBR signature is missing."
                   "\r\n         Creating signature..." );
    _mbr.mbrSignature[0] = 0x55;
    _mbr.mbrSignature[1] = 0xAA;
    result = FALSE;
  }


  // Check partitions
  for( i = 0; i < 4; ++i )
  {
    // Only check occupied partitions
    if( _mbr.partition[i].typeId != 0x00 &&
        _mbr.partition[i].nSectors > 0 )
    {
      // Check if partition ends past the end of disk
      if( _mbr.partition[i].startLSector + _mbr.partition[i].nSectors - 1 >
          (unsigned long)_mediaData.nCylinders * _mediaData.nHeads * _mediaData.nSectors - 1 )
      {
        helper_printf( "\r\nError: Partition %i ends past end of disk.", i );
        result = FALSE;
      }

      // Logical and physical addresses fit?
      if( ( fdiskPhysicToLogic( _mbr.partition[i].startCylinder,
                                (unsigned short)_mbr.partition[i].startHead,
                                (unsigned short)_mbr.partition[i].startSector ) !=
            _mbr.partition[i].startLSector )
          ||
          ( fdiskPhysicToLogic( _mbr.partition[i].endCylinder,
                                (unsigned short)_mbr.partition[i].endHead,
                                (unsigned short)_mbr.partition[i].endSector ) !=
            _mbr.partition[i].startLSector + _mbr.partition[i].nSectors - 1 ) )
      {
        helper_printf( "\r\nWarning: Physical and logical addresses of partition %i don't match."
                       "\r\n         This might be alright. You should check it though.", i );
        result = FALSE;
      }

      // Check if partition overlaps with other partition
      for( j = i; j < 4; ++j )
      {
        // Of course the partition poverlaps with itself
        // and of course all free partitions overlap
        if( j != i && _mbr.partition[j].typeId != 0 &&
            _mbr.partition[j].nSectors != 0 )
        {
        if( ( _mbr.partition[i].startLSector > _mbr.partition[j].startLSector &&
              _mbr.partition[i].startLSector < _mbr.partition[j].startLSector + _mbr.partition[j].nSectors - 1 )
              ||
              ( _mbr.partition[j].startLSector > _mbr.partition[i].startLSector &&
                _mbr.partition[j].startLSector < _mbr.partition[i].startLSector + _mbr.partition[i].nSectors - 1 ) )
          {
            // Partitions i and j overlap
            helper_printf( "\r\nWarning: Partitions %i and %i overlap.", i, j );
            result = FALSE;
          }
        }
      }
    }
  }

  // Sort partition entries
  for( i = 3; i > 0; --i )
  {
    for( j = i - 1; j >= 0; --j )
    {
      // Partitions that have smaller sectorOffset first
      // Partitions with sectorOffset 0 last
      if( _mbr.partition[i].startLSector != 0 &&
          ( _mbr.partition[j].startLSector > _mbr.partition[i].startLSector ||
            _mbr.partition[j].startLSector == 0 ) )
      {
        memcpy( _buffer, &( _mbr.partition[i] ), sizeof( TPartition ) );
        memcpy( &( _mbr.partition[i] ), &( _mbr.partition[j] ),
                sizeof( TPartition ) );
        memcpy( &( _mbr.partition[j] ), _buffer, sizeof( TPartition ) );
      }
    }
  }

  return result;
}



/******************************************************************************
* fdiskWriteMbr()
*
* Writes the MBR structure in the global variable _mbr to disk.
*
* Parameters:
*  None
*
* Result
*  TRUE  - On success.
*  FALSE - If Master Boot Record could not be written to disk.
******************************************************************************/
FdiskBool fdiskWriteMbr( void )
{
  unsigned short i,       // Counter variable
                 offset;  // Offset of partition table entry


  // First check for correct interface version
  if( !fdiskCheckInterfaceVersion() )
    return FALSE;

  // Fill _buffer[]
  memcpy( _buffer, &_mbr, 0x01BE );

  // Partition table entries
  for( i = 0; i < 4; ++i )
  {
    offset = 0x01BE + i * PARTITION_ENTRY_SIZE;

    _buffer[offset+0] = _mbr.partition[i].bootState;

    _buffer[offset+1] = _mbr.partition[i].startHead;
    _buffer[offset+2] = _mbr.partition[i].startSector & 0x3F;
    _buffer[offset+3] = (unsigned char)( _mbr.partition[i].startCylinder & 0x00FF );
    _buffer[offset+2] |= (unsigned char)( ( _mbr.partition[i].startCylinder & 0x0300 ) >> 2 );

    _buffer[offset+4] = _mbr.partition[i].typeId;

    _buffer[offset+5] = _mbr.partition[i].endHead;
    if( _mbr.partition[i].endSector > 0x3F ||
        _mbr.partition[i].endCylinder > 0x03FF )
    {
      return FALSE;
    }
    _buffer[offset+6] = _mbr.partition[i].endSector & 0x3F;
    _buffer[offset+7] = (unsigned char)( _mbr.partition[i].endCylinder & 0x00FF );
    _buffer[offset+6] |= (unsigned char)( ( _mbr.partition[i].endCylinder & 0x0300 ) >> 2 );

    _buffer[offset+8] = (unsigned char)( ( _mbr.partition[i].startLSector & 0x000000FFl ) >> 0 );
    _buffer[offset+9] = (unsigned char)( ( _mbr.partition[i].startLSector & 0x0000FF00l ) >> 8 );
    _buffer[offset+10] = (unsigned char)( ( _mbr.partition[i].startLSector & 0x00FF0000l ) >> 16 );
    _buffer[offset+11] = (unsigned char)( ( _mbr.partition[i].startLSector & 0xFF000000l ) >> 24 );

    _buffer[offset+12] = (unsigned char)( ( _mbr.partition[i].nSectors & 0x000000FFl ) >> 0 );
    _buffer[offset+13] = (unsigned char)( ( _mbr.partition[i].nSectors & 0x0000FF00l ) >> 8 );
    _buffer[offset+14] = (unsigned char)( ( _mbr.partition[i].nSectors & 0x00FF0000l ) >> 16 );
    _buffer[offset+15] = (unsigned char)( ( _mbr.partition[i].nSectors & 0xFF000000l ) >> 24 );
  }

  _buffer[510] = _mbr.mbrSignature[0];
  _buffer[511] = _mbr.mbrSignature[1];


  // Write first sector of disk
  if( !extlibWrite( _drive, 0, 1, _buffer ) )
  {
    return FALSE;
  }

  return TRUE;
}



/******************************************************************************
* fdiskClearMbr()
*
* Clears the MBR structure in the global variable _mbr and writes it to disk.
*
* Parameters:
*  None
*
* Result
*  TRUE  - On success.
*  FALSE - If Master Boot Record could not be written to disk.
******************************************************************************/
FdiskBool fdiskClearMbr( void )
{
  // Clear whole structure
  memset( &_mbr, 0x00, sizeof( TMasterBootRecord ) );

  // Set executable marker
  _mbr.mbrSignature[0] = 0x55;
  _mbr.mbrSignature[1] = 0xAA;

  // Write to disk
  return fdiskWriteMbr();
}



/******************************************************************************
* fdiskGetFirstFreePartition()
*
* Searches for a free partition entry in the MBR structure in the global
* variable _mbr.
*
* Parameters:
*  None
*
* Result
*  Nnumber of the first free partition entry, where 1 is the first partition and
*  4 is the last one. Returns 0 if there's no free partition left.
******************************************************************************/
unsigned char fdiskGetFirstFreePartition( void )
{
  unsigned short i;  // Counter variable


  for( i = 0; i < 4; ++i )
  {
    if( _mbr.partition[i].typeId == 0x00 ||
        _mbr.partition[i].nSectors == 0 )  // Partition inactive
    {
      return i + 1;
    }
  }

  return 0;
}



/******************************************************************************
* fdiskGetFirstFreeSector()
*
* Searches for the first sector which is not occupied by a partition.
* Starts searching behind the last active partition.
*
* Parameters:
*  None
*
* Result
*  Logical number of the sector or 0 if there is no free sector. 0 can't be a
*  valid return value for it's the start sector of the MBR.
******************************************************************************/
unsigned long fdiskGetFirstFreeSector( void )
{
  unsigned short i;             // Number of first free partition
  unsigned short cylinder,
                 head,
                 sector;         // Physical address


  i = fdiskGetFirstFreePartition();

  switch( i = fdiskGetFirstFreePartition() )
  {
    // First free partition is partition 1
    // => The whole disk is free
    case 1:
             fdiskLogicToPhysic( 1, &cylinder, &head, &sector );
             break;

    // First free partition is 2, 3 or 4
    // => First free sector follows the last occupied sector
    case 2:
    case 3:
    case 4:
             fdiskLogicToPhysic( _mbr.partition[i-2].startLSector + _mbr.partition[i-2].nSectors,
                                 &cylinder, &head, &sector );
             break;
             
    // There's no free partition
    // => There's no free sector
    case 0:
    default:
             return 0;
  }

  // Round up
  if( head != 0 || sector != 1 )
  {
    ++head;
    sector = 1;
  }
  
  // Check if new address is still on disk
  if( fdiskPhysicToLogic( cylinder, head, sector ) >=
      (unsigned long)_mediaData.nCylinders * _mediaData.nHeads * _mediaData.nSectors )
  {
    return 0;
  }
  else
  {
    return fdiskPhysicToLogic( cylinder, head, sector );
  }
}



/******************************************************************************
* fdiskGetLastFreeSector()
*
* Searches for the last sector which is not occupied by a partition.
* Starts searching behind the last active partition. So if there is a free
* sector, the last free sector will be the last one of the disk.
*
* Parameters:
*  None
*
* Result
*  Logical number of the sector or 0 if there is no free sector. 0 can't be a
*  valid return value for it's the start sector of the MBR.
******************************************************************************/
unsigned long fdiskGetLastFreeSector( void )
{
  // Check if there's any free sector
  if( fdiskGetFirstFreeSector() )
  {
    return ( (unsigned long)_mediaData.nCylinders * _mediaData.nHeads * _mediaData.nSectors ) - 1;
  }

  return 0;
}



/******************************************************************************
* fdiskCreatePartition()
*
* Creates a new partition.
*
* Parameters:
*  nSectors - Number of sectors in new partition
*  typeId   - Type of new  partition
*  bootable - TRUE if the new partition is to be bootable
*
* Result
*  Number of the new partition where 1 is the first partition and 4 the last
*  one. Returns 0 if a new partition could not be created.
******************************************************************************/
unsigned short fdiskCreatePartition( unsigned long nSectors,
                                     unsigned char typeId,
                                     FdiskBool bootable )
{
  unsigned short i;            // Number of the new partition
  TPartition partition;        // Safety copy of partition
  unsigned long startLSector;  // Logical start sector
  unsigned short startCylinder,
                 startHead,
                 startSector,
                 endCylinder,
                 endHead,
                 endSector;     // Temporary CHS values.


  // Get number of new partition
  i = fdiskGetFirstFreePartition();

  // Check if the paartition number is valid
  if( !i )
  {
    helper_strout( "\r\nError: All partitions are already occupied." );
    return 0;
  }

  // Check size
  if( fdiskGetLastFreeSector() - fdiskGetFirstFreeSector() + 1 < nSectors )
  {
    // Not enough free sectors
    helper_strout( "\r\nError: There is no free space left on disk." );
    return 0;
  }

  if( nSectors < _mediaData.nSectors )
  {
   // Partition too small
   helper_printf( "\r\nError: Partitions smaller than %u KB are not allowed.", _mediaData.nSectors / 2 );
   return 0;
  }

  // Check typeId
  if( typeId != FDISK_T_FAT12 &&
      typeId != FDISK_T_FAT16S &&
      typeId != FDISK_T_FAT16L &&
      typeId != FDISK_T_FAT32 &&
      typeId != FDISK_T_FAT32_LBA &&
      typeId != FDISK_T_FAT16L_LBA )
  {
    helper_printf( "\r\nError: Partitions of type 0x%02X are not supported.", typeId );
    return 0;
  }

  // Round up start address
  startLSector = fdiskGetFirstFreeSector();

  // Save copy of partiton entry
  memcpy( &partition, &_mbr.partition[i-1], sizeof( TPartition ) );

  // Write partition entry
  _mbr.partition[i-1].bootState = ( bootable ) ? 0x80 : 0x00;

  fdiskLogicToPhysic( startLSector,
                      &startCylinder,
                      &startHead,
                      &startSector );

  _mbr.partition[i-1].typeId = typeId;

  fdiskLogicToPhysic( startLSector + nSectors - 1,
                      &endCylinder,
                      &endHead,
                      &endSector );

  _mbr.partition[i-1].startLSector = startLSector;
  _mbr.partition[i-1].nSectors = nSectors;


  // Reduce values for CHS to their partition table maximums if necessary;
  // Partitions that are larger than what can be specified with correct CHS
  // values, will get the maximum CHS value and the correct logical values.
  if( startCylinder > FDISK_MAX_N_CYLINDERS ||
      startHead > FDISK_MAX_N_HEADS ||
      startSector > FDISK_MAX_N_SECTORS ||
      endCylinder > FDISK_MAX_N_CYLINDERS ||
      endHead > FDISK_MAX_N_HEADS ||
      endSector > FDISK_MAX_N_SECTORS )
  {
    _mbr.partition[i-1].startCylinder = FDISK_MAX_N_CYLINDERS;
    _mbr.partition[i-1].startHead = FDISK_MAX_N_HEADS;
    _mbr.partition[i-1].startSector = FDISK_MAX_N_SECTORS;
    _mbr.partition[i-1].endCylinder = FDISK_MAX_N_CYLINDERS;
    _mbr.partition[i-1].endHead = FDISK_MAX_N_HEADS;
    _mbr.partition[i-1].endSector = FDISK_MAX_N_SECTORS;
  }
  else
  {
    _mbr.partition[i-1].startCylinder = startCylinder;
    _mbr.partition[i-1].startHead = startHead;
    _mbr.partition[i-1].startSector = startSector;
    _mbr.partition[i-1].endCylinder = endCylinder;
    _mbr.partition[i-1].endHead = endHead;
    _mbr.partition[i-1].endSector = endSector;
  }


  // Write new MBR
  if( !fdiskWriteMbr() )
  {
    // Recover safety copy
    memcpy( &_mbr.partition[i-1], &partition, sizeof( TPartition ) );
    return 0;
  }

  // Clear boot block of new partition
  memset( _buffer, 0x00, 512 );
  if( !extlibWrite( _drive, _mbr.partition[i-1].startLSector, 1, _buffer ) )
    helper_strout( "\r\nWarning: Bootblock of new partition could not be cleared." );

  return i;
}



/******************************************************************************
* fdiskDeletePartition()
*
* Deletes the partition with the specified number.
*
* Parameters:
*  partitionNumber - Number of the patition to be deleted
*
* Result:
*  TRUE  - On success.
*  FALSE - If the partition could not be deleted.
******************************************************************************/
FdiskBool fdiskDeletePartition( unsigned char partitionNumber )
{
  TMasterBootRecord mbr;  // Safety copy of _mbr


  // Save copy of partiton entry
  memcpy( &mbr, &_mbr, sizeof( TMasterBootRecord ) );

  // Delete partition entry
  memset( &( _mbr.partition[partitionNumber-1] ), 0x00, sizeof( TPartition ) );

  // Sort partition entries
  fdiskCheckMbr();

  // write new MBR
  if( !fdiskWriteMbr() )
  {
    // Recover safety copy
    memcpy( &_mbr, &mbr, sizeof( TMasterBootRecord ) );
    return FALSE;
  }

  return TRUE;
}



/******************************************************************************
* fdiskGetPartitionData()
*
* Fills the given partition structure with the data from the specified partition.
*
* Parameters:
*  partitionNumber - Number of partition to get data from
*  partition       - Pointer to structure of type TPartition
*
* Result:
*  TRUE  - On success.
*  FALSE - If the partition number is invalid or the partition is free.
******************************************************************************/
FdiskBool fdiskGetPartitionData( unsigned char partitionNumber,
                                 TPartition *partition )
{
  // Check if partition number is valid
  if( partitionNumber < 1 || partitionNumber > 4 ||
      _mbr.partition[partitionNumber-1].typeId == 0x00 ||
      _mbr.partition[partitionNumber-1].nSectors == 0 )
  {
    return FALSE;
  }

  // Copy partition data
  memcpy( partition, &_mbr.partition[partitionNumber-1], sizeof( TPartition ) );

  return TRUE;
}



/******************************************************************************
* fdiskSetPartitionData()
*
* Changes the partition data and writes the new MBR to disk.
*
* Parameters:
*  partitionNumber - Number of partition
*  partition       - Pointer to structure of type TPartition
*
* Result:
*  TRUE  - On success.
*  FALSE - If MBR could not be written to disk or number of partition is invalid.
******************************************************************************/
FdiskBool fdiskSetPartitionData( unsigned char partitionNumber,
                                 TPartition *partition )
{
  TMasterBootRecord mbr;  // Safety copy of _mbr

  
  // Check if partition number is valid
  if( partitionNumber < 1 || partitionNumber > 4 ||
      _mbr.partition[partitionNumber-1].typeId == 0x00 ||
      _mbr.partition[partitionNumber-1].nSectors == 0 )
  {
    return FALSE;
  }

  // Save copy of partiton entry
  memcpy( &mbr, &_mbr, sizeof( TMasterBootRecord ) );

  // Set partition data
  memcpy( &_mbr.partition[partitionNumber-1], partition, sizeof( TPartition ) );

  // Check new MBR before writing it to disk
  if( !fdiskCheckMbr() )
  {
    // Recover safety copy
    memcpy( &_mbr, &mbr, sizeof( TMasterBootRecord ) );
    return FALSE;
  }

  // Write new MBR to disk
  if( !fdiskWriteMbr() )
  {
    // Recover safety copy
    memcpy( &_mbr, &mbr, sizeof( TMasterBootRecord ) );
    return FALSE;
  }

  return TRUE;
}



// End of file

